import java.util.*;
import java.sql.*;


public class ConnectionPool
{

  //nome driver jdbc
  private String driver = null;
  //nome url del db
  private String url = null;
  //numero iniziale di connessioni
  private int size = 0;
  //numero massimo di connessioni consentite in pi rispetto al valore iniziale
  //per comodit, questo valore pu essere espresso in percentuale
  private int pctIncrease = 0;
  //nome utente
  private String user = "";
  //password
  private String password = "";
  //vettore di connessioni JDBC
  private Vector pool = null;
  //DRIVER
  Driver drv = null;
  //questo parametro indica per quanto tempo una connessione pu rimanere viva
  //serve se per qualche motivo non riusciamo a rilasciare una connessione
  private long maxTimeInUse = 10000;


  /**
   * Questo costruttore  vuoto: se lo si usa, bisogna invocare i vari metodi set<p>
   * che impostano i vari dati.<br>
   */
  public ConnectionPool()
  {
  	
  }

  /**
   * questo costruttore imposta sin dall'inizio tutti i valori significativi.<br>
   */
  public ConnectionPool(String url, String driver, String user, String password,int size, int pctIncrease, long maxTimeInUse)
  {
    this.url = url;
    this.user = user;
    this.driver = driver;
    this.size = size;
    this.pctIncrease = pctIncrease;
    this.maxTimeInUse = maxTimeInUse;
  }

  //imposta il valore del driver
  public void setDriver(String path)
  {
    driver = (path==null?driver:path);
  }

  //legge il nome del driver
  public String getDriver()
  {
    return driver;
  }


  //imposta il nome della url del database
  public void setURL(String value)
  {
    url = (value==null?url:value);
  }

  //legge il nome della url
  public String getURL()
  {
    return url;
  }
  
  
  //imposta la massima percentuale di espansione del pool
  public void setPctIncrease(int pct)
  {
    pctIncrease = pct;
  }

  
  public String getPctIncrease()
  {
    return url;
  }


  //imposta il numero iniziale di connessioni
  public void setSize(int value)
  {
    size = (value > -1?value:size);
  }


  //Ottiene il numero iniziale di connessioni
  public int getSize()
  {
   return size;
  }

  //imposta il nome utente
  public void setUser(String aUser)
  {
    user = (aUser==null?user:aUser);
  }


  //legge il nome utente
  public String getUser()
  {
    return user;
  }
  
   //imposta il nome utente
  public void setPassword(String aPassword)
  {
    password = (aPassword==null?password:aPassword);
  }


  //legge il nome utente
  public String getPassword()
  {
    return password;
  }


  //imposta maxTimeInUse
  public void setMaxTimeInUse(long l)
  {
    maxTimeInUse = (l == 0?10000:l);
  }

  //ottiene maxTimeInUse
  public long getMaxTimeInUse()
  {
    return maxTimeInUse;
  }



  // Crea e restituisce una connessione

  private Connection createConnection() throws Exception
  {

   Connection con = null;

   con = DriverManager.getConnection(url,user,password);
  // System.out.println("Connessione in createConnection: "+con);
   con.setAutoCommit(false);
   return con;

  }

  /**
  *  il metodo inizializza il pool
  */
  public synchronized void initializePool() throws Exception
  {
    //test di tutte le variabili necessarie alla connessione
    if(driver == null)
    {
    	throw new Exception("Il nome del driver jdbc non  stato specificato");
    }

    if(url == null)
    {
    	throw new Exception("La URL del Database o la sorgente di dati non  stata specificata");
    }

    if(size < 1)
    {
    	throw new Exception("La dimensione del pool non pu essere < 1");
    }


    //creazione delle connessioni
    try
    {
    	// carica il file di classe del driver
    	drv = (Driver)Class.forName(driver).newInstance();
        DriverManager.registerDriver(drv);

    	// le connessioni vengono inzialmente create in base al numero size
    	for(int i = 0;i < size; i++)
    	{
    	  Connection con = createConnection();
    	  if(con != null)
    	  {
    	    //l'oggetto PoolableConnection deve contenere la connessione JDBC reale
    	    PoolableConnection pcon = new PoolableConnection(con);
    	    //la nuova connessione viene aggiunta al pool
    	    addConnection(pcon);

    	  }
    	}

    }
    catch(Exception e)
    {
      System.out.println("eccezione in initializeConnection: "+e.getMessage());
      throw new Exception(e.getMessage());
    }

  }


  //Aggiunge l'oggetto PoolableConnection al pool
  private void addConnection(PoolableConnection pcon)
  {
    //se il pool  null, crea un nuovo vettore, con dimensione iniziale 'size'
    if(pool == null)
    {
      pool = new Vector(size);
    }

    // aggiunge l'oggetto PoolableConnection al Vettore
    pool.addElement(pcon);
  }



  // Rilascia la connessione corrente
  public synchronized void releaseConnection(Connection con)
  {
    //questo flag vine impostato ogni volta che una connessione viene rilasciata
    //se non viene impostato true, vuol dire che il metodo non  riuscito a rilasciare
    //la connessione, quindi un thread separato si occuper di farlo rileggendo
    //il vettore maxTimeInUse/1000 volte.
    boolean foundAndKilled = false;
    //trova l'oggettoPoolableConnection
    for(int i = 0;i < pool.size(); i++)
    {
    	PoolableConnection pcon = (PoolableConnection)pool.elementAt(i);
    	// verifica se la connessione  corretta
    	if(pcon.getConnection() == con)
    	{
    	  System.out.println("sto rilasciando la connessione "+i);
    	  //imposta l'attributo inuse a false,
    	  //il che rende la connessione i disponibile
    	  pcon.setLocked(false);
          foundAndKilled = true;
    	  break;
    	}
    }
    if(!foundAndKilled)
    {
        System.out.println("dentro foundAndKilled");
	new ControlThread(maxTimeInUse,pool);
    }
  }


  // Trova una connessione disponibile
  public synchronized Connection getConnection() throws Exception
  {
    
    try{
    PoolableConnection pcon = null;
    // cerca una connessione disponibile
    int i = 0;
    for(i = 0; i < pool.size(); i++)
    {
    	pcon = (PoolableConnection)pool.elementAt(i);
    	if(!pcon.isLocked())
    	{
    	  pcon.setLocked(true);
    	  //System.out.println("pcon.getConnection()="+pcon.getConnection());
    	  return pcon.getConnection();
    	}
    }

    // Se non trova una connessione libera ne aggiunge una nuova se ha ancora spazio per farlo
    boolean expand = (size+(size/100*pctIncrease) < pool.size());
    if(expand)
    {
        return expandPool();
    }
    else
     return null;
    }
    catch(Exception e)
    {
        return null;
    } 
  }


  private Connection expandPool() throws Exception
  {
    
    PoolableConnection pcon = null;
    try
    {
      	// Crea una nuova connessione JDBC
      	Connection con = createConnection();
      	pcon = new PoolableConnection(con);
      	//imposta la connessione come in uso
      	pcon.setLocked(true);
      	//aggiunge il nuovo oggetto PoolableConnection al pool
      	pool.addElement(pcon);

    }
    catch(Exception e)
    {
    	System.err.println(e.getMessage());
    	throw new Exception(e.getMessage());
    }
    //restituisce la nuova connessione
    return pcon.getConnection();
  
  }


  ///prima di chiudere il pool, lo svuota
  public synchronized void emptyPool() throws SQLException
  {
     //itera sull'intero pool chiudendo le connessioni JDBC
     for(int i = 0;i < pool.size(); i++)
     {
     	PoolableConnection pcon = (PoolableConnection)pool.elementAt(i);
     	// se la connessione non  in uso, la chiude
     	if(!pcon.isLocked())
     	{
     	  pcon.close();
     	}
     	else
     	{
          try
          {
     	    //se la connessione  ancora in uso, attende maxTimeInUse e forza la chiusura
     	    java.lang.Thread.sleep(maxTimeInUse);
     	    pcon.close();
     	  }
     	  catch(InterruptedException ie)
     	  {
     	  	 System.out.println(ie.getMessage());
     	  }

     	}
     }
  }

}